"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  linode:
    driver: linode
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    endpoint_url: https://<zone_id>.linodeobjects.com

Possible zone IDs are the following:

- ``us-east-1``: Newark, NJ
- ``eu-central-1``: Frankfurt, DE
- ``ap-south-1``: Singapore, SG

.. _boto3: https://github.com/boto/boto3
"""
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Linode S3 Driver"""
    id = 'linode'

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.kwargs['endpoint_url'], bucket_id, name)
        return url



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  scalewey:
    driver: scaleway
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    region_name: <region_name>

Possible ``<region_name>`` are the following:

- ``fr-par``: Paris, France
- ``nl-ams``: Amsterdam, Netherland

.. _boto3: https://github.com/boto/boto3
"""
from urllib.parse import urljoin
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Scaleway S3 Driver"""
    id = 'scaleway'
    endpoint_url = 'https://s3.fr-par.scw.cloud'
    endpoint_urls = {
        'fr-par': 'https://s3.fr-par.scw.cloud',
        'nl-ams': 'https://s3-nl-ams.scw.cloud',
    }
    default_kwargs = {
        'endpoint_url': endpoint_url,
    }

    def get_custom_kwargs(self, kwargs):
        if 'region_name' in kwargs:
            endpoint_url = self.endpoint_urls.get(kwargs['region_name'])
            self.endpoint_url = endpoint_url
        kwargs['endpoint_url'] = self.endpoint_url
        return kwargs

    def get_url(self, bucket_id, name, **kwargs):
        url = urljoin(self.endpoint_url, '%s/%s' % (bucket_id, name))
        return url



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  google:
    driver: google_s3
    aws_access_key_id: <_ak>
    aws_secret_access_key: <your_sk>
    project_id: <project_id>

.. _boto3: https://github.com/boto/boto3
"""
from urllib.parse import urljoin
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Google Storage S3 Driver"""
    id = 'google_s3'
    endpoint_url = 'https://storage.googleapis.com'
    default_acl = 'public-read-write'
    default_object_acl = 'public-read'
    default_kwargs = {
        'endpoint_url': endpoint_url,
        'region_name': 'auto',
    }

    def _get_create_request_params(self, name, acl, **kwargs):
        params = super()._get_create_request_params(name, acl, **kwargs)
        if self.kwargs['region_name'] != 'auto':
            params['CreateBucketConfiguration'] = {
                'LocationConstraint': self.kwargs['region_name']
            }
        return params

    def get_url(self, bucket_id, name, **kwargs):
        url = urljoin(self.endpoint_url, '%s/%s' % (bucket_id, name))
        return url



"""
.. note::
  This driver requires `oss2`_.

`Object Storage Service`_ from `Alibaba Cloud`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  alibaba:
    driver: alibaba
    access_key_id: <your_ak>
    access_key_secret: <your_sk>
    endpoint: https://oss-<region_id>.aliyuncs.com

.. _oss2: https://github.com/aliyun/aliyun-oss-python-sdk
.. _`Object Storage Service`: https://www.alibabacloud.com/product/oss
.. _`Alibaba Cloud`: https://www.alibabacloud.com/
"""
import concurrent.futures
from functools import wraps
from urllib.parse import urlparse
import oss2
from os_benchmark.drivers import base, errors


class AliSession(oss2.Session):
    def __init__(self, raw_session):
        self.session = raw_session


def handle_request(method):
    @wraps(method)
    def _handle_request(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except oss2.exceptions.ServerError as err:
            if err.status == 403:
                self.logger.warning("Did you activate OSS in the web console?")
                raise errors.DriverPermissionError(err)
            raise
        except oss2.exceptions.RequestError as err:
            if err.status == -2:
                raise errors.DriverConnectionError(err.exception.args[0].reason.args[0])
            raise
    return _handle_request


class Driver(base.RequestsMixin, base.BaseDriver):
    """Alibaba OSS Driver"""
    id = 'alibaba'

    @property
    def auth(self):
        if not hasattr(self, '_auth'):
            self._auth = oss2.Auth(self.kwargs['access_key_id'], self.kwargs['access_key_secret'])
        return self._auth

    @property
    def service(self):
        if not hasattr(self, '_service'):
            self._service = oss2.Service(self.auth, self.kwargs['endpoint'])
        return self._service

    @property
    def ali_session(self):
        if not hasattr(self, '_ali_session'):
            self._ali_session = AliSession(self.session)
        return self._ali_session

    def _get_bucket(self, name):
        bucket = oss2.Bucket(
            auth=self.auth,
            endpoint=self.kwargs['endpoint'],
            bucket_name=name,
            session=self.ali_session,
            connect_timeout=self.connect_timeout,
        )
        return bucket

    @handle_request
    def list_buckets(self, **kwargs):
        raw_buckets = oss2.BucketIterator(self.service)
        buckets = [{'id': b.name} for b in raw_buckets]
        return buckets

    @handle_request
    def create_bucket(self, name, acl='public-read', **kwargs):
        bucket = self._get_bucket(name)
        bucket.create_bucket(
            permission=acl,
            input=oss2.models.BucketCreateConfig('Standard'),
        )
        return {'id': name}

    @handle_request
    def clean_multipart(self, bucket_id, **kwargs):
        bucket = self._get_bucket(bucket_id)
        parts = bucket.list_multipart_uploads().upload_list
        for part in parts:
            self.logger.debug('Aborting multipart %s/%s %s', bucket_id, part.key, part.upload_id)
            bucket.abort_multipart_upload(
                key=part.key,
                upload_id=part.upload_id,
            )

    @handle_request
    def delete_bucket(self, bucket_id, clean_multipart=True, **kwargs):
        bucket = self._get_bucket(bucket_id)
        if clean_multipart:
            self.clean_multipart(bucket_id)

        try:
            bucket.delete_bucket()
        except oss2.exceptions.BucketNotEmpty as err:
            raise errors.DriverNonEmptyBucketError(err)

    @handle_request
    def list_objects(self, bucket_id, **kwargs):
        bucket = self._get_bucket(bucket_id)
        raw_objects = bucket.list_objects().object_list
        objects = [o.key for o in raw_objects]
        return objects

    def _simple_upload(self, bucket_id, name, content, acl='public-read'):
        bucket = self._get_bucket(bucket_id)
        bucket.put_object(name, content)

    def _multipart_upload(self, bucket_id, name, content, acl='public-read', multipart_chunksize=None, max_concurrency=None):
        bucket = self._get_bucket(bucket_id)
        multipart_chunksize = multipart_chunksize or base.MULTIPART_CHUNKSIZE
        max_concurrency = max_concurrency or base.MAX_CONCURRENCY

        content_size = content.size
        part_id = 1
        offset = 0
        parts = []
        upload_id = bucket.init_multipart_upload(name).upload_id

        def _upload(part_id, offset):
            self.logger.debug('Uploading %s part %s', name, part_id)
            result = bucket.upload_part(name, upload_id, part_id, oss2.SizedFileAdapter(content, multipart_chunksize))
            self.logger.debug('Done %s part %s', name, part_id)
            parts.append(oss2.models.PartInfo(part_id, result.etag, size=multipart_chunksize, part_crc=result.crc))

        pool_kwargs = {'max_workers': max_concurrency}
        with concurrent.futures.ThreadPoolExecutor(**pool_kwargs) as executor:
            futures = []
            while offset < content_size:
                chunk_size = min(multipart_chunksize, content_size - offset)
                result = executor.submit(_upload, part_id=part_id, offset=offset)
                futures.append(result)
                part_id += 1
                offset += multipart_chunksize
            for future in futures:
                future.result()

        bucket.complete_multipart_upload(name, upload_id, parts)

    @handle_request
    def upload(self, bucket_id, name, content, acl='public-read',
               multipart_threshold=None, multipart_chunksize=None,
               max_concurrency=None,
               **kwargs):
        multipart_threshold = multipart_threshold or base.MULTIPART_THRESHOLD
        if content.size > multipart_threshold:
            self._multipart_upload(bucket_id, name, content, acl, multipart_chunksize, max_concurrency)
        else:
            self._simple_upload(bucket_id, name, content, acl)
        return {'name': name}

    @handle_request
    def delete_object(self, bucket_id, name, **kwargs):
        bucket = self._get_bucket(bucket_id)
        bucket.delete_object(name)

    def get_url(self, bucket_id, name, **kwargs):
        bucket = self._get_bucket(bucket_id)
        hostname = urlparse(self.kwargs['endpoint']).netloc
        url = 'https://%s.%s/%s' % (bucket_id, hostname, name)
        return url



"""
.. note::
  This driver requires `boto3`_ and `ibm-cos-sdk`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  ibm_s3:
    driver: ibm
    ibm_api_key_id: <your_api_key>
    ibm_service_instance_id: <your_cos_id>
    endpoint_url: <endpoint_url>

``ibm_api_key_id``: IBM API key
``ibm_service_instance_id``: COS ID, ie ``crn:v1:bluemix:public:cloud-object-storage:global:a/c067b1bc8f094716bcd8e3acca9d0230:c36aev98-23ad-4b1d-a63b-b2a7d7b7ecde::``

.. _boto3: https://github.com/boto/boto3
.. _ibm_boto3: https://github.com/IBM/ibm-cos-sdk-python
"""
from urllib.parse import urljoin
import ibm_boto3
import ibm_botocore
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """IBM Cloud Object Storage Driver"""
    id = 'ibm'

    @property
    def s3(self):
        if not hasattr(self, '_s3'):
            kwargs = {
                'ibm_auth_endpoint': "https://iam.cloud.ibm.com/identity/token",
            }
            kwargs.update(self.kwargs)
            config = {'signature_version': "oauth"}
            config.update(self.kwargs.get('config') or {})
            kwargs['config'] = ibm_botocore.client.Config(**config)
            self._s3 = ibm_boto3.resource("s3", **kwargs)
        return self._s3

    def get_url(self, bucket_id, name, **kwargs):
        url = urljoin(self.kwargs['endpoint_url'], '%s/%s' % (bucket_id, name))
        return url



"""Error module"""
class OsbError(Exception):
    """Base OS Benchmark error"""


class ConfigurationError(OsbError):
    """Error with configuration file"""


class InvalidHttpCode(OsbError):
    """Invalid HTTP response from server"""



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  backblaze:
    driver: backblaze_s3
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    endpoint_url: s3.<zone_id>.backblazeb2.com

Possible zone IDs are the following:

- ``us-west-001``
- ``us-west-002``
- ``eu-central-003``

.. _boto3: https://github.com/boto/boto3
"""
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Backblaze S3 Driver"""
    id = 'backblaze_s3'

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.kwargs['endpoint_url'], bucket_id, name)
        return url

    def delete_object(self, bucket_id, name, **kwargs):
        versions = self.s3.meta.client.list_object_versions(
            Bucket=bucket_id, KeyMarker=name).get('DeleteMarkers', [])
        obj = self.s3.Object(bucket_id, name)
        for version in versions:
            if version['Key'] == name:
                self.logger.debug('Remove %s version %s', name, version['VersionId'])
                obj.delete(VersionId=version['VersionId'])



"""
.. note::
  This driver requires `boto3`_.

Base S3 driver allowing usage of any S3-based storage.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  s3:
    driver: s3
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    region: eu-west-1

All parameters except ``driver`` will be passed to ``boto3.resource``.
"""
from functools import wraps

import botocore
import boto3
from boto3.s3.transfer import TransferConfig

from os_benchmark.drivers import base, errors


def handle_request(method):
    @wraps(method)
    def _handle_request(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except botocore.exceptions.ConnectionClosedError as err:
            raise errors.DriverConnectionError(err)
        except botocore.exceptions.EndpointConnectionError as err:
            raise errors.DriverConnectionError(err)
        except botocore.exceptions.ReadTimeoutError as err:
            raise errors.DriverConnectionError(err)
        except botocore.exceptions.ClientError as err:
            code = err.response['Error']['Code']

            if 'Message' not in err.response['Error']:
                raise errors.DriverConnectionError(err)

            msg = err.response['Error']['Message']
            if code == '504':
                raise errors.DriverConnectionError(err)
            if code == 'ServiceUnavailable':
                raise errors.DriverConnectionError(err)
            if code == 'InvalidAccessKeyId':
                msg += " (endpoint: %s)" % self.s3.meta.client._endpoint.host
                raise errors.DriverAuthenticationError(msg)
            raise
    return _handle_request


class Driver(base.RequestsMixin, base.BaseDriver):
    id = 's3'
    default_acl = 'public-read'
    default_object_acl = 'public-read'
    default_kwargs = {}
    default_config = {}
    _default_config = {
        'user_agent': base.USER_AGENT,
        'retries': {'max_attempts': 0},
        'connect_timeout': base.CONNECT_TIMEOUT,
        'read_timeout': base.READ_TIMEOUT,
        'parameter_validation': False,
        # 'max_pool_connections': self.num_thread,
        # 'proxies': proxies,
    }

    @property
    def s3(self):
        if not hasattr(self, '_s3'):
            kwargs = self.kwargs.copy()
            kwargs.update(self.default_kwargs)
            kwargs.update(self.get_custom_kwargs(kwargs))

            config = self._default_config.copy()
            config.update(self.default_config)
            config.update(kwargs.pop('config', None) or {})
            if self.read_timeout is not None:
                config['read_timeout'] = self.read_timeout
            if self.connect_timeout is not None:
                config['connect_timeout'] = self.connect_timeout
            self.logger.debug("boto Config: %s", config)
            kwargs['config'] = botocore.client.Config(**config)

            self._s3 = boto3.resource('s3', **kwargs)
        return self._s3

    def get_custom_kwargs(self, kwargs):
        return kwargs

    @handle_request
    def list_buckets(self, **kwargs):
        raw_buckets = self.s3.buckets.all()
        buckets = [{'id': b.name} for b in raw_buckets]
        return buckets

    def _get_create_request_params(self, name, acl, **kwargs):
        return {
            'Bucket': name,
            'ACL': acl,
        }

    @handle_request
    def create_bucket(self, name, acl=None, **kwargs):
        acl = acl or self.default_acl
        params = self._get_create_request_params(name=name, acl=acl, **kwargs)
        self.logger.debug('Create bucket params: %s', params)
        bucket = self.s3.create_bucket(**params)
        return {'id': name}

    @handle_request
    def delete_bucket(self, bucket_id, **kwargs):
        bucket = self.s3.Bucket(bucket_id)
        try:
            bucket.delete()
        except botocore.exceptions.ClientError as err:
            code = err.response['Error']['Code']
            msg = err.response['Error']['Message']
            if code == 'NoSuchBucket':
                self.logger.debug(err)
                return
            if code == 'BucketNotEmpty':
                raise errors.DriverNonEmptyBucketError(msg)
            raise

    @handle_request
    def list_objects(self, bucket_id, **kwargs):
        bucket = self.s3.Bucket(bucket_id)
        try:
            objects = [o.key for o in bucket.objects.all()]
        except botocore.exceptions.ClientError as err:
            code = err.response['Error']['Code']
            msg = err.response['Error']['Message']
            if code == 'NoSuchBucket':
                raise errors.DriverBucketUnfoundError(msg)
            raise
        return objects

    @handle_request
    def upload(self, bucket_id, name, content, acl=None,
               multipart_threshold=None, multipart_chunksize=None,
               max_concurrency=None,
               **kwargs):
        acl = acl or self.default_object_acl
        extra = {'ACL': acl}
        multipart_threshold = multipart_threshold or base.MULTIPART_THRESHOLD
        multipart_chunksize = multipart_chunksize or base.MULTIPART_CHUNKSIZE
        max_concurrency = max_concurrency or base.MAX_CONCURRENCY

        transfer_config = TransferConfig(
            multipart_threshold=multipart_threshold,
            max_concurrency=max_concurrency,
            multipart_chunksize=multipart_chunksize,
        )
        try:
            self.s3.meta.client.upload_fileobj(
                Fileobj=content,
                Bucket=bucket_id,
                Key=name,
                ExtraArgs=extra,
                Config=transfer_config,
            )
        except botocore.exceptions.ClientError as err:
            code = err.response['Error']['Code']
            msg = err.response['Error']['Message']
            if code == 'NoSuchBucket':
                raise errors.DriverBucketUnfoundError(msg)
            elif code == 'AccessDenied':
                raise errors.DriverBucketUnfoundError(msg)
            raise
        return {'name': name}

    @handle_request
    def delete_object(self, bucket_id, name, **kwargs):
        obj = self.s3.Object(bucket_id, name)
        obj.delete()



"""
.. note::
  This driver requires `oci`_.

`Object Storage`_ from `Oracle Cloud`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  oracle:
    driver: oracle
    fingerprint:
    region: FRA
    tenancy:
    namespace:
    key_content: '-----BEGIN RSA PRIVATE KEY-----

    blablabla

    stillblala

    enoughblabla

    -----END RSA PRIVATE KEY-----
    '

.. _oci: https://github.com/aliyun/aliyun-oss-python-sdk
.. _`Object Storage`: https://www.alibabacloud.com/product/oss
.. _`Oracle Cloud`: https://www.alibabacloud.com/
"""
import concurrent.futures
from functools import wraps
import oci
from oci.object_storage import models
import requests
from requests.packages.urllib3.util.retry import Retry
from os_benchmark.drivers import base, errors

ACLS = {
    'public-read': 'ObjectRead',
}


class MultiPart:
    """Object simulating part from file-object for multipart-upload."""
    def __init__(self, file_object, size):
        self.file_object = file_object
        self.size = size
        self.offset = 0

    def read(self, chunksize=None):
        if self.offset >= self.size:
            return ''

        if (chunksize is None or chunksize < 0) or (chunksize + self.offset >= self.size):
            data = self.file_object.read(self.size - self.offset)
            self.offset = self.size
            return data

        self.offset += chunksize
        return self.file_object.read(chunksize)

    @property
    def len(self):
        return self.size


def handle_request(method):
    @wraps(method)
    def _handle_request(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except oci.exceptions.ServiceError as err:
            if err.code == 'NotAuthenticated':
                raise errors.DriverAuthenticationError(err)
            raise
        except requests.exceptions.ConnectionError as err:
            raise errors.DriverConnectionError(err)
    return _handle_request


class Driver(base.RequestsMixin, base.BaseDriver):
    """Oracle Cloud Infrastructure Driver"""
    id = 'oracle'

    @property
    def client(self):
        if not hasattr(self, '_client'):
            try:
                self._client = oci.object_storage.ObjectStorageClient(
                    config=self.kwargs,
                    timeout=self.read_timeout,
                    retry_strategy=oci.retry.NoneRetryStrategy(),
                )
            except oci.exceptions.InvalidConfig as err:
                raise errors.DriverConfigError(err)
            retry = Retry(total=0)
            timeout = (self.connect_timeout, self.read_timeout)
            adapter = base.HTTPAdapter(max_retries=retry, timeout=timeout)
            self._client.base_client.session.mount('http://', adapter)
            self._client.base_client.session.mount('https://', adapter)
        return self._client

    @handle_request
    def list_buckets(self, **kwargs):
        response = self.client.list_buckets(
            namespace_name=self.kwargs['namespace'],
            compartment_id=self.kwargs['tenancy'],
        )
        buckets = [{'id': b.name} for b in response.data]
        return buckets

    @handle_request
    def create_bucket(self, name, acl='public-read', **kwargs):
        acl = ACLS[acl]
        create_opts = models.CreateBucketDetails(
            name=name,
            compartment_id=self.kwargs['tenancy'],
            public_access_type=acl,
            storage_tier='Standard',
        )
        response = self.client.create_bucket(
            namespace_name=self.kwargs['namespace'],
            create_bucket_details=create_opts
        )
        return {'id': name}

    @handle_request
    def delete_bucket(self, bucket_id, **kwargs):
        try:
            self.client.delete_bucket(
                namespace_name=self.kwargs['namespace'],
                bucket_name=bucket_id,
            )
        except oci.exceptions.ServiceError as err:
            if err.code == 'BucketNotEmpty':
                raise errors.DriverNonEmptyBucketError(err)
            elif err.code == 'BucketNotFound':
                return
            raise

    @handle_request
    def list_objects(self, bucket_id, **kwargs):
        try:
            response = self.client.list_objects(
                namespace_name=self.kwargs['namespace'],
                bucket_name=bucket_id,
            )
        except oci.exceptions.ServiceError as err:
            if err.code == 'BucketNotFound':
                raise errors.DriverBucketUnfoundError(err)
            raise
        return [o.name for o in response.data.objects]

    def _multipart_upload(self, bucket_id, name, content, multipart_chunksize=None, max_concurrency=None):
        multipart_chunksize = multipart_chunksize or base.MULTIPART_CHUNKSIZE
        max_concurrency = max_concurrency or base.MAX_CONCURRENCY

        content_size = content.size
        part_id = 1
        offset = 0
        parts = []

        create_multipart_upload_details = models.CreateMultipartUploadDetails()
        upload_data = self.client.create_multipart_upload(
            namespace_name=self.kwargs['namespace'],
            bucket_name=bucket_id,
            create_multipart_upload_details=create_multipart_upload_details,
        )
        upload_id = upload_data.init_multipart_upload(name).upload_id

        def _upload(part_id, offset):
            self.logger.debug('Uploading %s part %s', name, part_id)
            part = MultiPart(content, multipart_chunksize)
            self.client.upload_part(
                namespace_name=self.kwargs['namespace'],
                bucket_name=bucket_id,
                object_name=name,
                upload_part_num=part_id,
                upload_part_body=part,
            )
            self.logger.debug('Done %s part %s', name, part_id)

        pool_kwargs = {'max_workers': max_concurrency}
        with concurrent.futures.ThreadPoolExecutor(**pool_kwargs) as executor:
            futures = []
            while offset < content_size:
                chunk_size = min(multipart_chunksize, content_size - offset)
                result = executor.submit(_upload, part_id=part_id, offset=offset)
                futures.append(result)
                part_id += 1
                offset += multipart_chunksize
            for future in futures:
                future.result()

        commit_multipart_upload_details = models.CommitMultipartUploadDetails()
        self.client.commit_multipart_upload(
            namespace_name=self.kwargs['namespace'],
            bucket_name=bucket_id,
            object_name=name,
            upload_id=upload_id,
            commit_multipart_upload_details=commit_multipart_upload_details,
        )

    def _simple_upload(self, bucket_id, name, content, acl='public-read'):
        response = self.client.put_object(
            namespace_name=self.kwargs['namespace'],
            bucket_name=bucket_id,
            object_name=name,
            put_object_body=content,
        )

    @handle_request
    def upload(self, bucket_id, name, content,
               multipart_threshold=None, multipart_chunksize=None,
               max_concurrency=None,
               **kwargs):
        multipart_threshold = multipart_threshold or base.MULTIPART_THRESHOLD
        self.logger.debug('Multipart: %s > %s -> %s', content.size, multipart_threshold, content.size > multipart_threshold)
        if content.size > multipart_threshold:
            self._multipart_upload(bucket_id, name, content, multipart_chunksize, max_concurrency)
        else:
            self._simple_upload(bucket_id, name, content)
        return {'name': name}

    @handle_request
    def delete_object(self, bucket_id, name, **kwargs):
        try:
            self.client.delete_object(
                namespace_name=self.kwargs['namespace'],
                bucket_name=bucket_id,
                object_name=name,
            )
        except oci.exceptions.ServiceError as err:
            if err.code in ('BucketNotFound', 'ObjectNotFound'):
                return
            raise

    def get_url(self, bucket_id, name, **kwargs):
        path = '/n/%s/b/%s/o/%s' % (
            self.kwargs['namespace'],
            bucket_id,
            name
        )
        url = '%s%s' % (self.client.base_client._endpoint, path)
        return url



"""
.. note::
  This driver requires `python-swiftclient`_.

`Object Storage`_ from `bluvalt`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  bluvalt:
    driver: bluvalt
    user: <your_tenant_id>:<your_username>
    key: <your_password>

.. _python-swiftclient: https://github.com/openstack/python-swiftclient
.. _`Object Storage`: https://cloud.bluvalt.com/#/application-marketplace/958/object-storage-958/1
.. _bluvalt: https://bluvalt.com/
"""

from os_benchmark.drivers import swift


class Driver(swift.Driver):
    """Bluvalt Swift Driver"""
    id = 'bluvalt'
    default_kwargs = {
        'authurl': 'https://api-object.bluvalt.com:8083/auth/v1.0',
    }

    @property
    def session_headers(self):
        if not hasattr(self, '_session_headers'):
            self.base_url, self.token = self.swift.get_auth()
            self._session_headers = {'X-Auth-Token': self.token}
        return self._session_headers

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.swift.url, bucket_id, name)
        return url



"""
Filesystem driver storing buckets in file-system with true directories.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  myFsrofile:
    driver: fs
    path: /tmp/osn/
"""
import os
import shutil
from os_benchmark.drivers import base


class Driver(base.BaseDriver):
    id = 'fs'

    def setup(self):
        os.makedirs(self.path)

    @property
    def path(self):
        if not hasattr(self, '_path'):
            self._path = self.kwargs['path']
        return self._path

    def _create_directory(self, path):
        os.makedirs(path)

    def _list_files(self, path):
        return os.listdir(path)

    def list_buckets(self, **kwargs):
        return [
            {'id': f} for f in self._list_files(self.path)
        ]

    def create_bucket(self, name, **kwargs):
        path = os.path.join(self.path, name)
        self._create_directory(path)
        return {'id': name}

    def delete_bucket(self, bucket_id):
        path = os.path.join(self.path, bucket_id)
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            self.logger.debug("Directory '%s' doesn't exist", path)

    def list_objects(self, bucket_id, **kwargs):
        path = os.path.join(self.path, bucket_id)
        try:
            self._create_directory(path)
        except FileExistsError:
            pass
        bucket_files = []
        for root, dirs, files in os.walk(path):
            for file_ in files:
                bucket_files.append(
                    file_
                )
        return bucket_files

    def upload(self, bucket_id, name, content, **kwargs):
        path = os.path.join(self.path, bucket_id, name)
        directory_path = os.path.dirname(path)
        try:
            self._create_directory(directory_path)
        except FileExistsError:
            pass

        with open(path, 'wb') as fd:
            shutil.copyfileobj(
                fsrc=content,
                fdst=fd,
            )
        return {'name': name}

    def get_url(self, bucket_id, name, **kwargs):
        path = os.path.join(self.path, bucket_id, name)
        url = 'file://%s' % path
        return url

    def download(self, url, block_size=2048, **kwargs):
        path = url.replace('file://', '')
        with open(path, 'rb') as fd:
            while fd.read(block_size):
                pass

    def delete_object(self, bucket_id, name, **kwargs):
        bucket_path = os.path.join(self.path, bucket_id)
        obj_path = os.path.join(bucket_id, name)
        abs_obj_path = os.path.join(self.path, obj_path)
        try:
            os.remove(abs_obj_path)
        except FileNotFoundError:
            self.logger.debug("File '%s' doesn't exist", abs_obj_path)
        except NotADirectoryError:
            self.logger.debug("'%s' is a directory", abs_obj_path)
        # Delete dirs if empty
        dirs = [i for i in obj_path.split('/') if i]
        bucket_i = dirs.index(bucket_id)+1
        for i in range(len(dirs))[::-1]:
            if i == bucket_i:
                break
            dir_path = '/'.join(dirs[:i])
            abs_dir_path = os.path.join(self.path, dir_path)
            try:
                files = self._list_files(abs_dir_path)
            except FileNotFoundError:
                self.logger.debug("File '%s' doesn't exist", obj_path)
            else:
                if not files:
                    break
            shutil.rmtree(abs_dir_path)



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  exoscale:
    driver: exoscale
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    endpoint_url: https://sos-<zone_id>.exo.io

Possible zone IDs are the following:

- ``de-fra-1``: Frankfurt, Germany
- ``ch-dk-2``: Zurich, Switzerland
- ``at-vie-1``: Vienna, Austria
- ``de-muc-1``: Munich, Germany
- ``ch-gva-2``: Geneva, Switzerland
- ``bg-sof-1``: Sofia, Bulgaria

.. _boto3: https://github.com/boto/boto3
"""
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Exoscale S3 Driver"""
    id = 'exoscale'

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.kwargs['endpoint_url'], bucket_id, name)
        return url



"""Driver errors module."""
from os_benchmark import errors as base


class DriverError(base.OsbError):
    """Base Driver error"""


class DriverConnectionError(DriverError):
    """Driver connection error"""


class DriverAuthenticationError(DriverError):
    """Driver authentication error"""


class DriverPermissionError(DriverError):
    """Driver permission error"""


class DriverConfigError(DriverError):
    """Driver configuration error"""


class DriverNonEmptyBucketError(DriverError):
    """Driver non-empty bucket error"""


class DriverBucketUnfoundError(DriverError):
    """Driver bucket unfound error"""


class DriverObjectUnfoundError(DriverError):
    """Driver object unfound error"""


class DriverStorageQuotaError(DriverError):
    """Driver service storage quota error"""



"""
.. note::
  This driver requires `python-swiftclient`_.

`Object Storage`_ from `OVHcloud`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  ovhcloud:
    driver: ovhcloud
    user: <your_tenant_id>:<your_username>
    key: <your_password>
    tenant_name: "<your tenant_name>"
    os_options: 
      tenant_id: <your_tenant_id>
      region_name: <region_id>
      project_name: <your_project_name>
      project_id: <you_project_id>

.. warning::
    `tenant_name` **must** be surrounded by quotes.

.. _python-swiftclient: https://github.com/openstack/python-swiftclient
.. _`Object Storage`: https://www.ovhcloud.com/en/public-cloud/object-storage/
.. _OVHcloud: https://www.ovhcloud.com/
"""

from os_benchmark.drivers import swift


class Driver(swift.Driver):
    """OVHcloud Swift Driver"""
    id = 'ovhcloud'
    default_kwargs = {
        'authurl': 'https://auth.cloud.ovh.net/v3/',
        'auth_version': 3,
    }

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.swift.url, bucket_id, name)
        return url



"""
Command-line management module.
"""
import os
import sys
import argparse
import json
import logging
import os_benchmark
from os_benchmark import utils, benchmarks, errors
from os_benchmark.drivers import errors as driver_errors

ACTIONS = (
    'help',
    'create-bucket',
    'list-buckets',
    'delete-bucket',
    'list-objects',
    'upload',
    'download',
    'delete-object',
    'clean-bucket',
    'clean',
    'time-upload',
    'time-download',
)
MULTIPART_THREHOLD = 64 * 2**20
MULTIPART_CHUNKSIZE = 8 * 2**20
MAX_CONCURRENCY = os.cpu_count() * 2

logger = logging.getLogger('osb')

def create_parser():
    """Create main parser"""
    parser = argparse.ArgumentParser(
        prog='os-benchmark',
        add_help=False,
    )
    parser.add_argument(
        '--config-file',
        required=False,
        help="Specify a configuration file to use."
    )
    parser.add_argument(
        '--config-raw',
        required=False,
        help="Provide a raw configuration as JSON instead of a stored file.",
    )
    parser.add_argument(
        '--config-name',
        required=False,
        help="Select a driver configuration to use."
    )
    parser.add_argument(
        '-C', '--connect-timeout',
        default=5, required=False, type=float,
        help="The time in seconds till a timeout is considered during TCP connection.",
    )
    parser.add_argument(
        '-R', '--read-timeout',
        default=10, required=False, type=float,
        help="The time in seconds till a timeout is considered durint HTTP read",
    )
    parser.add_argument(
        '-v', '--verbosity',
        default=0, required=False, type=int,
        choices=(0, 1, 2, 3),
        help="Verbosity level; 0=minimal output, 1=normal output 2=verbose output 3=still more",
    )
    parser.add_argument(
        '-i', '--noinput',
        default=False, action='store_true',
        help="Disable any prompt",
    )
    return parser


class Controller:
    """Helper for organise CLI work"""
    def __init__(self):
        self.parser = create_parser()
        self.subparsers = self.parser.add_subparsers(help="Sub-command", dest='action')

        action_subparsers = {}
        main_action = base_num_args = None
        for action in ACTIONS:
            action_subparsers[action] = self.subparsers.add_parser(action)
            if action in sys.argv:
                main_action = action
                base_num_args = sys.argv.index(action) + 1
        self.main_args = self.parser.parse_known_args(sys.argv[1:base_num_args])[0]
        main_action = self.main_args.action or 'help'
        self.subparser = action_subparsers[main_action]
        self.action = main_action.replace('-', '_')
        # Logs
        self.verbosity = 30 - (self.main_args.verbosity * 10)
        self.logger = logger
        console_handler = logging.StreamHandler()
        self.logger.addHandler(console_handler)
        self.logger.setLevel(self.verbosity)
        # Get config
        if self.main_args.config_raw:
            config = json.loads(self.main_args.config_raw)
        else:
            try:
                config = utils.get_driver_config(
                    config_name=self.main_args.config_name,
                    config_file=self.main_args.config_file,
                )
            except errors.ConfigurationError as err:
                self.logger.error(err)
                self.help()
        config['read_timeout'] = self.main_args.read_timeout
        config['connect_timeout'] = self.main_args.connect_timeout
        # Get driver
        self.driver = utils.get_driver(config)

    def run(self):
        func = getattr(self, self.action)
        result = func()
        return result

    def help(self):
        self.parser.print_help()
        self.parser.exit()

    def create_bucket(self):
        self.subparser.add_argument('--name', required=False)
        self.subparser.add_argument('--storage-class', required=False)
        parsed_args = self.parser.parse_known_args()[0]

        name = parsed_args.name or utils.get_random_name()
        bucket = self.driver.create_bucket(
            name=name,
            storage_class=parsed_args.storage_class,
        )
        return bucket

    def delete_bucket(self):
        self.subparser.add_argument('bucket_id')
        self.subparser.add_argument('--delete-files', action='store_true')
        parsed_args = self.parser.parse_known_args()[0]

        if parsed_args.delete_files:
            try:
                self.driver.clean_bucket(bucket_id=parsed_args.bucket_id)
            except driver_errors.DriverBucketUnfoundError:
                return
        self.driver.delete_bucket(
            bucket_id=parsed_args.bucket_id,
        )

    def list_buckets(self):
        parsed_args = self.parser.parse_known_args()[0]
        buckets = self.driver.list_buckets()
        for bucket in buckets:
            print(bucket['id'])

    def upload(self):
        self.subparser.add_argument('--bucket-id')
        self.subparser.add_argument('--storage-class', required=False)
        self.subparser.add_argument('--name', required=False)
        content_group = self.subparser.add_mutually_exclusive_group()
        content_group.add_argument('--content', type=argparse.FileType('rb'), required=False)
        content_group.add_argument('--content-size', type=int, required=False)
        content_group.add_argument('--', '--from-stdin', default=False, action='store_true', dest='from_stdin')
        self.subparser.add_argument('--multipart-threshold', type=int, default=MULTIPART_THREHOLD)
        self.subparser.add_argument('--multipart-chunksize', type=int, default=MULTIPART_CHUNKSIZE)
        self.subparser.add_argument('--max-concurrency', type=int, default=MAX_CONCURRENCY)
        parsed_args = self.parser.parse_known_args()[0]

        name = parsed_args.name or utils.get_random_name()
        if parsed_args.from_stdin:
            content = sys.stdin
        elif parsed_args.content is not None:
            content = parsed_args.content
        elif parsed_args.content_size:
            content = utils.get_random_content(parsed_args.content_size)
        else:
            msg = "No input file given."
            raise errors.OsbError(msg)

        obj = self.driver.upload(
            bucket_id=parsed_args.bucket_id,
            storage_class=parsed_args.storage_class,
            name=name,
            content=content,
            multipart_threshold=parsed_args.multipart_threshold,
            multipart_chunksize=parsed_args.multipart_chunksize,
            max_concurrency=parsed_args.max_concurrency,
        )
        return obj

    def download(self):
        self.subparser.add_argument('--bucket-id')
        self.subparser.add_argument('--name')
        parsed_args = self.parser.parse_known_args()[0]
        url = self.driver.get_url(
            bucket_id=parsed_args.bucket_id,
            name=parsed_args.name,
        )
        self.logger.debug('URL: %s', url)
        fd = self.driver.download(url)
        print(fd.read())


    def list_objects(self):
        self.subparser.add_argument('bucket_id')
        self.subparser.add_argument('--url', action='store_true')
        parsed_args = self.parser.parse_known_args()[0]
        try:
            objects = self.driver.list_objects(
                bucket_id=parsed_args.bucket_id,
            )
        except driver_errors.DriverBucketUnfoundError as err:
            self.logger.warning(err.args[0])
            return
        for obj in objects:
            if parsed_args.url:
                url = self.driver.get_url(
                    bucket_id=parsed_args.bucket_id,
                    name=obj
                )
                print(obj, url)
            else:
                print(obj)

    def delete_object(self):
        self.subparser.add_argument('bucket_id')
        self.subparser.add_argument('name')
        parsed_args = self.parser.parse_known_args()[0]

        self.driver.delete_object(
            bucket_id=parsed_args.bucket_id,
            name=parsed_args.name,
        )

    def clean_bucket(self):
        self.subparser.add_argument('bucket_id')
        parsed_args = self.parser.parse_known_args()[0]

        if not self.main_args.noinput:
            print("You are going to clean entirely this bucket.")
            input("Press [ENTER] to continue\n")
        self.driver.clean_bucket(
            bucket_id=parsed_args.bucket_id,
        )

    def clean(self):
        parsed_args = self.parser.parse_known_args()[0]

        if not self.main_args.noinput:
            print("You are going to clean entirely this object storage.")
            input("Press [ENTER] to continue\n")
        self.driver.clean()

    def time_upload(self):
        self.subparser.add_argument('--storage-class', required=False)
        self.subparser.add_argument('--object-size', type=int, required=True)
        self.subparser.add_argument('--object-number', type=int, required=True)
        self.subparser.add_argument('--multipart-threshold', type=int, default=MULTIPART_THREHOLD)
        self.subparser.add_argument('--multipart-chunksize', type=int, default=MULTIPART_CHUNKSIZE)
        self.subparser.add_argument('--max-concurrency', type=int, default=MAX_CONCURRENCY)
        parsed_args = self.parser.parse_known_args()[0]

        benchmark = benchmarks.UploadBenchmark(self.driver)
        benchmark.set_params(
            storage_class=parsed_args.storage_class,
            object_size=parsed_args.object_size,
            object_number=parsed_args.object_number,
            multipart_threshold=parsed_args.multipart_threshold,
            multipart_chunksize=parsed_args.multipart_chunksize,
            max_concurrency=parsed_args.max_concurrency,
        )
        benchmark.setup()
        benchmark.run()
        benchmark.tear_down()
        stats = benchmark.make_stats()
        self.print_stats(stats)

    def time_download(self):
        self.subparser.add_argument('--storage-class', required=False)
        self.subparser.add_argument('--object-size', type=int, required=False)
        self.subparser.add_argument('--object-number', type=int, required=False)
        parsed_args = self.parser.parse_known_args()[0]

        benchmark = benchmarks.DownloadBenchmark(self.driver)
        benchmark.set_params(
            storage_class=parsed_args.storage_class,
            object_size=parsed_args.object_size,
            object_number=parsed_args.object_number,
        )
        benchmark.setup()
        benchmark.run()
        benchmark.tear_down()
        stats = benchmark.make_stats()
        self.print_stats(stats)

    def print_stats(self, stats):
        template = '%s\t\t%s'
        print(template % ('version', os_benchmark.__version__))
        for key, value in stats.items():
            if isinstance(value, float):
                value = round(value, 10)
                print('%s\t\t%f' % (key, value))
            else:
                print(template % (key, value))


def main():
    """Entry function"""
    try:
        controller = Controller()
        controller.run()
    except KeyboardInterrupt:
        print("Stopped by user")
        sys.exit(2)
    except errors.OsbError as err:
        if logger.level <= 0:
            raise
        logger.error(err)
        sys.exit(1)


if __name__ == '__main__':
    main()
    sys.exit(0)



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  otc:
    driver: otc
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>

.. _boto3: https://github.com/boto/boto3
"""
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """T-Systems Open Cloud Telekom S3 Driver"""
    id = 'otc'
    endpoint_url = 'https://obs.otc.t-systems.com'
    default_kwargs = {
        'endpoint_url': endpoint_url,
    }

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.default_kwargs['endpoint_url'], bucket_id, name)
        return url



"""
.. note::
  This driver requires `python-swiftclient`_.

Base `Openstack Swift`_ driver allowing usage of any Swift-based storage.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  swift_connection:
    driver: swift
    user: <your_username>
    key: <your_password>
    authurl: <auth_url>

All parameters except ``driver`` will be passed to ``swiftclient.Connection``.

.. _python-swiftclient: https://github.com/openstack/python-swiftclient
.. _`Openstack Swift`: https://github.com/openstack/swift
"""
import swiftclient
from swiftclient.service import SwiftService, SwiftUploadObject
from os_benchmark.drivers import base, errors


class Driver(base.RequestsMixin, base.BaseDriver):
    id = 'swift'
    default_kwargs = {}

    @property
    def swift(self):
        if not hasattr(self, '_swift'):
            kwargs = self.kwargs
            kwargs.update(self.default_kwargs)
            self._swift = swiftclient.Connection(**self.kwargs)
        return self._swift

    def setup(self, **kwargs):
        self.service_kwargs = kwargs.copy()
        self.service_kwargs.update(
            retries=0
        )

        if 'max_concurrency' in kwargs:
            self.service_kwargs.update(
                object_threads=kwargs['max_concurrency'],
                object_uu_threads=kwargs['max_concurrency'],
                segment_threads=kwargs['max_concurrency'],
            )
        if 'multipart_chunksize' in kwargs:
            self.service_kwargs['segment_size'] = kwargs['multipart_chunksize']
        swiftclient.service._default_global_options.update(self.service_kwargs)
        self.logger.debug("Swift default config: %s", self.service_kwargs)
        self.swift  # Avoid lazyness

    def list_buckets(self, **kwargs):
        _, raw_buckets = self.swift.get_account()
        buckets = [
            {'id': b['name']}
            for b in raw_buckets
        ]
        return buckets

    def create_bucket(self, name, acl='public-read', **kwargs):
        headers = {}
        if acl == 'public-read':
            headers['X-Container-Read'] = '.r:*'
        self.swift.put_container(name, headers=headers)
        return {'id': name}

    def delete_bucket(self, bucket_id, **kwargs):
        try:
            self.swift.delete_container(bucket_id)
        except swiftclient.ClientException as err:
            if err.http_status == 404:
                return
            if err.http_status == 409:
                raise errors.DriverNonEmptyBucketError(err.args[0])

    def list_objects(self, bucket_id, **kwargs):
        try:
            _, objs = self.swift.get_container(bucket_id)
        except swiftclient.ClientException as err:
            if err.http_status == 404:
                raise errors.DriverBucketUnfoundError(err.args[0])
            raise
        return [o['name'] for o in objs]

    def delete_object(self, bucket_id, name, **kwargs):
        try:
            self.swift.delete_object(bucket_id, name)
        except swiftclient.ClientException as err:
            if err.http_status == 404:
                return
            raise

    def upload(self, bucket_id, name, content, acl='public-read', **kwargs):
        self.swift.put_object(bucket_id, name, content)
        return {'name': name}



"""
.. note::
  This driver requires `b2sdk`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  backblaze:
    driver: backblaze
    application_key_id: <key_id>
    application_key: <key>

.. _b2sdk: https://github.com/Backblaze/b2-sdk-python
"""
from functools import wraps
import hashlib
from requests.packages.urllib3.util.retry import Retry
from b2sdk import v1 as b2
from b2sdk import api, bucket as bucket_, exception, utils
from b2sdk.transfer.outbound.upload_source import AbstractUploadSource
from os_benchmark.drivers import base, errors

ACLS = {
    'public-read': 'allPublic',
    'private': 'allPrivate',
}


class UploadSourceFileIo(AbstractUploadSource):
    def __init__(self, file):
        self.file = file

    def get_content_length(self):
        return self.file.size

    def get_content_sha1(self):
        self.file.seek(0)
        return hashlib.sha1(self.file.read()).hexdigest()

    def open(self):
        return self.file

    def seek(self, offset):
        self.file.seek(offset)

    def read(self, size=None):
        return self.file.read(size)


def handle_request(method):
    @wraps(method)
    def _handle_request(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except exception.B2ConnectionError as err:
            raise errors.DriverConnectionError(err)
        except exception.B2RequestTimeoutDuringUpload as err:
            raise errors.DriverConnectionError(err)
    return _handle_request


class Driver(base.RequestsMixin, base.BaseDriver):
    id = 'backblaze'

    @property
    def client(self):
        if not hasattr(self, '_client'):
            self._account_info = b2.InMemoryAccountInfo()
            self._client = api.B2Api(self._account_info)
            self.client.authorize_account(
                'production',
                self.kwargs['application_key_id'],
                self.kwargs['application_key']
            )
            retry = Retry(total=6)
            timeout = (self.connect_timeout, self.read_timeout)
            adapter = base.HTTPAdapter(max_retries=retry, timeout=timeout)
            self._client.raw_api.b2_http.session.mount('http://', adapter)
            self._client.raw_api.b2_http.session.mount('https://', adapter)
            self._client.raw_api.b2_http.TIMEOUT = timeout
        return self._client

    def list_buckets(self, **kwargs):
        buckets = self.client.list_buckets()
        return [{'id': c.id_} for c in buckets]

    @handle_request
    def create_bucket(self, name, acl='public-read', **kwargs):
        acl = ACLS.get(acl)
        bucket = self.client.create_bucket(
            name=name,
            bucket_type=acl,
        )
        return {
            'id': bucket.id_,
            'name': name,
        }

    def _get_bucket(self, bucket_id):
        bucket = self.client.get_bucket_by_id(bucket_id)
        return bucket

    def delete_bucket(self, bucket_id, **kwargs):
        bucket = self._get_bucket(bucket_id)
        try:
            unfinisheds = bucket.list_unfinished_large_files()
        except exception.NonExistentBucket:
            return
        for unfinished in unfinisheds:
            bucket.cancel_large_file(unfinished.file_id)

        try:
            self.client.delete_bucket(bucket)
        except exception.BadRequest as err:
            if err.code == 'cannot_delete_non_empty_bucket':
                raise errors.DriverNonEmptyBucketError(err.message)
            raise

    def list_objects(self, bucket_id, **kwargs):
        bucket = self._get_bucket(bucket_id)
        objs = bucket.ls()
        return [o.file_name for o, _ in objs]

    def _simple_upload(self, bucket_id, name, upload_source):
        bucket = self._get_bucket(bucket_id)
        content_length = upload_source.get_content_length()
        bucket.api.session.upload_file(
            bucket_id=bucket_id,
            file_name=name,
            content_length=content_length,
            content_type='application/octet-stream',
            content_sha1='do_not_verify',
            file_infos={},
            data_stream=upload_source,
        )

    def _multipart_upload(self, bucket_id, name, content, multipart_chunksize=None, max_concurrency=None):
        bucket = self._get_bucket(bucket_id)

        def _upload(part_id, offset, content, file_id):
            self.logger.debug('Uploading %s part %s', name, part_id)
            part = base.MultiPart(content, multipart_chunksize)
            upload_source = UploadSourceFileIo(content)
            result = bucket.api.session.upload_part(
                file_id=file_id,
                part_number=part_id,
                sha1_sum='do_not_verify',
                content_length=content.size,
                input_stream=upload_source,
            )
            self.logger.debug('Done %s part %s', name, part_id)
            return result

        result = bucket.api.session.start_large_file(
            bucket_id=bucket_id,
            file_name=name,
            content_type='application/octet-stream',
            file_info={}
        )
        file_id = result['fileId']

        uploader = base.MultiPartUploader(
            content=content,
            multipart_chunksize=multipart_chunksize,
            max_concurrency=max_concurrency,
            extre_uplood_kwargs={
                'file_id': file_id,
            }
        )
        parts = uploader.run(_upload)

        bucket.api.session.finish_large_file(
            file_id=file_id,
            part_sha1_array=[p['contentSha1'].replace('unverified:', '')  for p in parts],
        )

    @handle_request
    def upload(self, bucket_id, name, content, max_concurrency=None,
               multipart_chunksize=None, multipart_threshold=None,
               validate_content=False, **kwargs):
        multipart_threshold = multipart_threshold or base.MULTIPART_THRESHOLD
        upload_source = UploadSourceFileIo(content)

        try:
            if content.size > multipart_threshold:
                self._multipart_upload(bucket_id, name, content, multipart_chunksize, max_concurrency)
            else:
                self._simple_upload(bucket_id, name, upload_source)
        except exception.StorageCapExceeded as err:
            raise errors.DriverStorageQuotaError(err)
        return {'name': name}

    @handle_request
    def delete_object(self, bucket_id, name, **kwargs):
        bucket = self._get_bucket(bucket_id)
        versions = bucket.list_file_versions(name)
        for version in versions:
            try:
                bucket.delete_file_version(
                    version.id_,
                    name,
                )
            except exception.FileNotPresent:
                return

    def get_url(self, bucket_id, name, bucket_name, **kwargs):
        bucket = self._get_bucket(bucket_id)
        bucket.name = bucket_name
        url = bucket.get_download_url(name)
        return url



"""
In-memory driver keeping buckets and objects in RAM.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  myRamProfile:
    driver: ram
"""
import io
from shutil import copyfileobj
from urllib.parse import urlparse
from os_benchmark.drivers import base, errors

class Driver(base.BaseDriver):
    id = 'ram'

    def __init__(self):
        self.buckets = {}

    def list_buckets(self, **kwargs):
        return [{'id': b} for b in self.buckets]

    def delete_bucket(self, bucket_id, **kwargs):
        self.buckets.pop(bucket_id, None)

    def create_bucket(self, name, **kwargs):
        self.buckets[name] = {}
        return {'id': name}

    def list_objects(self, bucket_id, **kwargs):
        if bucket_id not in self.buckets:
            raise errors.DriverBucketUnfoundError("Bucket not found")
        return list(self.buckets[bucket_id])

    def upload(self, bucket_id, name, content, acl='public-read', **kwargs):
        if bucket_id not in self.buckets:
            raise errors.DriverBucketUnfoundError("Bucket not found")

        stored = io.BytesIO()
        copyfileobj(content, stored)
        stored.seek(0)

        self.buckets[bucket_id][name] = stored
        return {'name': name}

    def delete_object(self, bucket_id, name, **kwargs):
        if bucket_id not in self.buckets:
            raise errors.DriverBucketUnfoundError("Bucket not found")
        self.buckets[bucket_id].pop(name, None)

    def get_url(self, bucket_id, name, **kwargs):
        return 'ram://%s/%s' % (bucket_id, name)

    def download(self, url, block_size=65536, **kwargs):
        parsed_url = urlparse(url)

        bucket_id = parsed_url.netloc
        if bucket_id not in self.buckets:
            raise errors.DriverBucketUnfoundError("Bucket not found")

        obj_name = parsed_url.path[1:]
        if obj_name not in self.buckets[bucket_id]:
            raise errors.DriverObjectUnfoundError("Object not found")

        fd = self.buckets[bucket_id][obj_name]
        for chunk in fd.read(block_size):
            pass
        self.buckets[bucket_id][obj_name].seek(0)



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  aws:
    driver: aws
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    endpoint_url: https://s3.<region_id>.amazonaws.com
    region_name: <region_id>

.. _boto3: https://github.com/boto/boto3
"""
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """AWS S3 Driver"""
    id = 'aws'

    def _get_create_request_params(self, name, acl, **kwargs):
        params = super()._get_create_request_params(name, acl, **kwargs)
        params['CreateBucketConfiguration'] = {
            'LocationConstraint': self.kwargs['region_name']
        }
        return params

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.kwargs['endpoint_url'], bucket_id, name)
        return url



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  oracle:
    driver: oracle_s3
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>
    endpoint_url: https://<namespace>.compat.objectstorage.<zone_id>.oraclecloud.com

.. _boto3: https://github.com/boto/boto3
"""
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Oracle Cloud S3 Driver"""
    id = 'oracle_s3'
    default_acl = 'ObjectRead'

    def get_url(self, bucket_id, name, **kwargs):
        url = '%s/%s/%s' % (self.kwargs['endpoint_url'], bucket_id, name)
        return url



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  bso:
    driver: bso
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>

.. _boto3: https://github.com/boto/boto3
"""
from urllib.parse import urljoin
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """BSO-ST Driver"""
    id = 'bso'
    endpoint_url = 'https://eu.bso.st'
    default_kwargs = {
        'endpoint_url': endpoint_url,
    }
    default_config = {
        'signature_version': 's3v4',
    }

    def get_url(self, bucket_id, name, **kwargs):
        url = urljoin(self.endpoint_url, '%s/%s' % (bucket_id, name))
        return url



import os
import logging
import time

import yaml
from faker import Faker
import randomio

from os_benchmark import errors
from os_benchmark.drivers import utils as driver_utils

logger = logging.getLogger('osb.utils')
faker = Faker()


def get_config_file(config_file=None):
    """
    Get full configuration
    """
    if 'OSB_CONFIG_FILE' in os.environ:
        files = [os.environ['OSB_CONFIG_FILE']]
    elif config_file is not None:
        files = [config_file]
    else:
        files = ['~/.osb.yml', '/etc/osb.yml']

    configs = None
    for filename in files:
        filename = os.path.expanduser(filename)
        try:
            with open(filename) as fd:
                configs = yaml.full_load(fd)
                break
        except FileNotFoundError:
            continue

    logger.info("Use config file '%s'", filename)
    if not configs:
        msg = "No configuration file found."
        raise errors.ConfigurationError(msg)
    return configs


def get_driver_config(config_name=None, config_file=None):
    """
    Get driver configuration as dict
    """
    configs = get_config_file(config_file=config_file)

    if len(configs) == 1 and config_name is None:
        config_name = list(configs.keys())[0]
        logger.debug("Use the single driver config '%s'", config_name)

    if config_name and config_name not in configs:
        msg = "'%s' config found." % config_name
        raise errors.ConfigurationError(msg)

    if config_name is None:
        msg = 'Unknown configuration, please specify one in %s' % (
            ', '.join([c for c in configs])
        )
        raise errors.ConfigurationError(msg)

    return configs[config_name]


def get_driver(config):
    """Get configured driver"""
    key = config.pop('driver')
    driver_class = driver_utils.get_driver_class(key)
    logger.debug("Driver configured with '%s'", config)
    driver = driver_class(**config)
    return driver


def get_random_name(size=30):
    """Creates a random name"""
    name = faker.user_name()
    while len(name) < size:
        name += faker.user_name()
    return name[:size]


def get_random_content(size):
    """Creates a random fileobj"""
    return randomio.FileGenerator(size)


def timeit(func, *args, **kwargs):
    """Time a function"""
    start = time.time()
    output = func(*args, **kwargs)
    end = time.time()
    elapsed = end - start
    return elapsed, output



import importlib


def get_driver_class(key):
    """Get driver from its key"""
    module_path = 'os_benchmark.drivers.%s' % key
    module = importlib.import_module(module_path)
    driver_class = getattr(module, 'Driver')
    return driver_class



"""
Base Driver class module.
"""
import logging
import requests
import concurrent.futures
from requests.adapters import HTTPAdapter as BaseHTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from os_benchmark.drivers import errors

USER_AGENT = 'os-benchmark'
MULTIPART_THRESHOLD = 64*2**20
MULTIPART_CHUNKSIZE = 64*2**20
MAX_CONCURRENCY = 10
CONNECT_TIMEOUT = 3
READ_TIMEOUT = 1


class MultiPart:
    """Object simulating part from file-object for multipart-upload."""
    def __init__(self, file_object, size):
        self.file_object = file_object
        self.size = size
        self.offset = 0

    def read(self, chunksize=None):
        if self.offset >= self.size:
            return ''

        if (chunksize is None or chunksize < 0) or (chunksize + self.offset >= self.size):
            data = self.file_object.read(self.size - self.offset)
            self.offset = self.size
            return data

        self.offset += chunksize
        return self.file_object.read(chunksize)

    @property
    def len(self):
        return self.size


class MultiPartUploader:
    """Helper creating a thread pool and splitting file in several parts."""
    def __init__(self, content, max_concurrency=None, multipart_chunksize=None, extre_uplood_kwargs=None):
        self.content = content
        self.max_concurrency = max_concurrency or MAX_CONCURRENCY
        self.extre_uplood_kwargs = extre_uplood_kwargs or {}
        self.multipart_chunksize = multipart_chunksize or MULTIPART_CHUNKSIZE

    def run(self, upload_func):
        content_length = self.content.size
        part_id = 1
        offset = 0
        futures = []

        pool_kwargs = {'max_workers': self.max_concurrency}
        with concurrent.futures.ThreadPoolExecutor(**pool_kwargs) as executor:
            while offset < content_length:
                chunk_size = min(self.multipart_chunksize, content_length - offset)
                result = executor.submit(
                    upload_func,
                    content=self.content,
                    part_id=part_id,
                    offset=offset,
                    **self.extre_uplood_kwargs,
                )

                futures.append(result)
                part_id += 1
                offset += self.multipart_chunksize

            for future in futures:
                future.result()
        return futures


class BaseDriver:
    """Base Driver class"""
    id = None
    read_timeout = READ_TIMEOUT
    connect_timeout = CONNECT_TIMEOUT

    def __init__(self, read_timeout=None, connect_timeout=None, **kwargs):
        self.read_timeout = read_timeout or self.read_timeout
        self.connect_timeout = connect_timeout or self.connect_timeout
        self.kwargs = self._validate_kwargs(kwargs)
        self.logger = logging.getLogger('osb.driver')

    def setup(self, **kwargs):
        """Initialiaze driver before benchmark"""

    def _validate_kwargs(self, kwargs):
        """Ensure kwargs passed to __init__ are correct."""
        return kwargs

    def list_buckets(self, **kwargs):
        """List all buckets"""
        raise NotImplementedError()

    def create_bucket(self, name, **kwargs):
        """Create a bucket"""
        raise NotImplementedError()

    def delete_bucket(self, bucket_id):
        """Delete a bucket"""
        raise NotImplementedError()

    def list_objects(self, bucket_id, **kwargs):
        """List objects from a bucket"""
        raise NotImplementedError()

    def upload(self, bucket_id, name, content, **kwargs):
        """Upload an object into a bucket"""
        raise NotImplementedError()

    def get_url(self, bucket_id, name, **kwargs):
        """Get object URL"""
        raise NotImplementedError()

    def download(self, url, block_size=65536, **kwargs):
        """Download object from URL"""
        raise NotImplementedError()

    def delete_object(self, bucket_id, name, **kwargs):
        """Delete object from a bucket"""
        raise NotImplementedError()

    def clean_bucket(self, bucket_id, delete_bucket=True):
        """Delete all object from a bucket"""
        try:
            self.logger.debug("Listing all objects from %s", bucket_id)
            objects = self.list_objects(bucket_id=bucket_id)
        except errors.DriverBucketUnfoundError as err:
            self.logger.debug(err)
            return
        for obj in objects:
            self.logger.info("Deleting object %s/%s", bucket_id, obj)
            self.delete_object(bucket_id=bucket_id, name=obj)
        if delete_bucket:
            self.delete_bucket(bucket_id=bucket_id)

    def clean(self):
        """Delete all buckets and all object"""
        self.logger.debug("Listing all buckets")
        for bucket in self.list_buckets():
            self.logger.info("Deleting bucket %s", bucket['id'])
            self.clean_bucket(bucket_id=bucket['id'], delete_bucket=True)


class HTTPAdapter(BaseHTTPAdapter):
    def __init__(self, timeout=None, *args, **kwargs):
        self.timeout = 3 if timeout is None else timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)

class RequestsMixin:
    """Mixin providing a HTTTP Session"""
    session_headers = {}

    @property
    def session(self):
        if not hasattr(self, '_session'):
            self._session = requests.Session()
            self._session.headers = self.session_headers.copy()
            retry = Retry(total=0)
            timeout = (self.connect_timeout, self.read_timeout)
            adapter = HTTPAdapter(max_retries=retry, timeout=timeout)
            self._session.mount('http://', adapter)
            self._session.mount('https://', adapter)
        return self._session

    def download(self, url, block_size=65536, **kwargs):
        self.logger.debug('GET %s', url)
        try:
            with self.session.get(url, stream=True) as response:
                if response.status_code != 200:
                    msg = '%s %s' % (url, response.content)
                    raise errors.base.InvalidHttpCode(msg, response.status_code)
                for chunk in response.iter_content(chunk_size=block_size):
                    pass
        except requests.exceptions.ConnectionError as err:
            raise errors.DriverConnectionError(err.args[0])



"""Benchmark modules"""
import logging
import statistics
from os_benchmark import utils, errors
from os_benchmark.drivers import errors as driver_errors


class BaseBenchmark:
    """Base Benchmark class"""
    def __init__(self, driver):
        self.driver = driver
        self.logger = logging.getLogger('osb')
        self.params = {}

    def set_params(self, **kwargs):
        """Set test parameters"""
        self.params.update(kwargs)

    def setup(self):
        """Build benchmark environment"""

    def tear_down(self):
        """Destroy benchmark environment"""

    def run(self, **kwargs):
        """Run benchmark"""
        raise NotImplementedError()

    def make_stats(self):
        """Compute statistics as dict"""
        return {}


class UploadBenchmark(BaseBenchmark):
    """Time objects uploading"""
    def setup(self):
        self.timings = []
        self.objects = []
        self.errors = []
        self.driver.setup(**self.params)

        bucket_name = utils.get_random_name()
        self.storage_class = self.params.get('storage_class')
        self.logger.debug("Creating bucket '%s'", bucket_name)
        self.bucket = self.driver.create_bucket(
            name=bucket_name,
            storage_class=self.storage_class,
        )

    def run(self, **kwargs):
        def upload_files():
            for i in range(self.params['object_number']):
                name = utils.get_random_name()
                content = utils.get_random_content(self.params['object_size'])

                self.logger.debug("Uploading object '%s'", name)
                try:
                    elapsed, obj = utils.timeit(
                        self.driver.upload,
                        bucket_id=self.bucket['id'],
                        storage_class=self.storage_class,
                        name=name,
                        content=content,
                        multipart_threshold=self.params['multipart_threshold'],
                        multipart_chunksize=self.params['multipart_chunksize'],
                        max_concurrency=self.params['max_concurrency'],
                    )
                    self.timings.append(elapsed)
                    self.objects.append(obj)
                except driver_errors.DriverConnectionError as err:
                    self.logger.error(err)
                    self.errors.append(err)


        self.total_time = utils.timeit(upload_files)[0]

    def tear_down(self):
        self.driver.clean_bucket(bucket_id=self.bucket['id'])

    def make_stats(self):
        count = len(self.timings)
        error_count = len(self.errors)
        size = self.params['object_size']
        total_size = count * size
        test_time = sum(self.timings)
        rate = (count/test_time) if test_time else 0
        bw = (total_size/test_time/2**20) if test_time else 0
        stats = {
            'operation': 'upload',
            'ops': count,
            'time': self.total_time,
            'bw': bw,
            'rate': rate,
            'object_size': size,
            'object_number': self.params['object_number'],
            'multipart_threshold': self.params['multipart_threshold'],
            'multipart_chunksize': self.params['multipart_chunksize'],
            'max_concurrency': self.params['max_concurrency'],
            'total_size': total_size,
            'test_time': test_time,
            'errors': error_count,
            'driver': self.driver.id,
            'read_timeout': self.driver.read_timeout,
            'connect_timeout': self.driver.connect_timeout,
        }
        if count > 1:
            stats.update({
                'avg': statistics.mean(self.timings),
                'stddev': statistics.stdev(self.timings),
                'med': statistics.median(self.timings),
                'min': min(self.timings),
                'max': max(self.timings),
            })
        return stats


class DownloadBenchmark(BaseBenchmark):
    """Time objects downloading"""

    def setup(self):
        self.timings = []
        self.errors = []
        self.objects = []
        bucket_name = utils.get_random_name()
        self.logger.debug("Creating bucket '%s'", bucket_name)
        self.storage_class = self.params.get('storage_class')
        self.bucket = self.driver.create_bucket(
            name=bucket_name,
            storage_class=self.storage_class,
        )
        self.bucket_id = self.bucket['id']
        for i in range(self.params['object_number']):
            name = utils.get_random_name()
            content = utils.get_random_content(self.params['object_size'])

            self.logger.debug("Uploading object '%s'", name)
            try:
                obj = self.driver.upload(
                    bucket_id=self.bucket_id,
                    storage_class=self.storage_class,
                    name=name,
                    content=content,
                )
            except driver_errors.DriverError as err:
                self.logger.warning("Error during file uploading, tearing down the environment.")
                self.tear_down()
                raise
            self.objects.append(obj)
        self.urls = [
            self.driver.get_url(
                bucket_id=self.bucket_id,
                name=obj['name'],
                bucket_name=self.bucket.get('name', self.bucket_id),
            )
            for obj in self.objects
        ]

    def run(self, **kwargs):
        def download_objets(urls):
            for url in urls:
                try:
                    elapsed = utils.timeit(
                        self.driver.download,
                        url=url,
                    )[0]
                    self.timings.append(elapsed)
                except errors.InvalidHttpCode as err:
                    self.errors.append(err)

        self.total_time = utils.timeit(download_objets, urls=self.urls)[0]

    def tear_down(self):
        self.driver.clean_bucket(bucket_id=self.bucket['id'])

    def make_stats(self):
        count = len(self.timings)
        error_count = len(self.errors)
        size = self.params['object_size']
        total_size = count * size
        test_time = sum(self.timings)
        bw = (total_size/test_time/2**20) if test_time else 0
        rate = (count/test_time) if test_time else 0
        stats = {
            'operation': 'download',
            'ops': count,
            'time': self.total_time,
            'bw': bw,
            'rate': rate,
            'object_size': size,
            'object_number': self.params['object_number'],
            'max_concurrency': 1,
            'multipart_threshold': 0,
            'multipart_chunksize': 0,
            'total_size': total_size,
            'test_time': test_time,
            'errors': error_count,
            'driver': self.driver.id,
            'read_timeout': self.driver.read_timeout,
            'connect_timeout': self.driver.connect_timeout,
        }
        if count > 1:
            stats.update({
                'avg': statistics.mean(self.timings),
                'stddev': statistics.stdev(self.timings),
                'med': statistics.median(self.timings),
                'min': min(self.timings),
                'max': max(self.timings),
            })
        if error_count:
            error_codes = set([e for e in self.errors])
            stats.update({'error_count_%s' % e.args[1]: 0 for e in self.errors})
            for err in self.errors:
                key = 'error_count_%s' % err.args[1]
                stats[key] += 1
        return stats



"""
.. note::
  This driver requires `boto3`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  wasabi:
    driver: wasabi
    aws_access_key_id: <your_ak>
    aws_secret_access_key: <your_sk>

Possible region IDs are the following:

- ``us-east-1``: US East 1 (N. Virginia)
- ``us-east-2``: US East 2 (N. Virginia)
- ``us-west-1``: US West 1 (Oregon)
- ``eu-central-1``: EU Central 1 (Amsterdam)

.. _boto3: https://github.com/boto/boto3
"""
from urllib.parse import urljoin
from os_benchmark.drivers import s3


class Driver(s3.Driver):
    """Wasabi S3 Driver"""
    id = 'wasabi'
    endpoint_url = 'https://s3.wasabisys.com'
    endpoint_urls = {
        'us-east-1': 'https://s3.us-east-1.wasabisys.com',
        'us-east-2': 'https://s3.us-east-2.wasabisys.com',
        'us-west-1': 'https://s3.us-west-1.wasabisys.com',
        'eu-central-1': 'https://s3.eu-central-1.wasabisys.com',
    }

    default_kwargs = {
        'endpoint_url': endpoint_url,
    }

    def get_custom_kwargs(self, kwargs):
        if 'region_name' in kwargs:
            endpoint_url = self.endpoint_urls.get(kwargs['region_name'])
            kwargs['endpoint_url'] = endpoint_url
            self.endpoint_url = endpoint_url
        return kwargs

    def get_url(self, bucket_id, name, **kwargs):
        url = urljoin(self.endpoint_url, '%s/%s' % (bucket_id, name))
        return url



import os
import tempfile
import time
from unittest import TestCase
import yaml
from os_benchmark import utils, errors


def create_config_file(config, filename=None):
    filename = filename or os.path.expanduser('~/.osb.yml')
    with open(filename, 'w') as fd:
        yaml.dump(config, stream=fd)


class GetConfigFileTest(TestCase):
    def test_environ(self):
        filename = tempfile.mktemp()
        create_config_file({'foo': {'driver': 'ram'}}, filename)
        os.environ['OSB_CONFIG_FILE'] = filename
        configs = utils.get_config_file()
        del os.environ['OSB_CONFIG_FILE']

    def test_specified(self):
        filename = tempfile.mktemp()
        create_config_file({'foo': {'driver': 'ram'}}, filename)
        configs = utils.get_config_file(filename)

    def test_default(self):
        filename = os.path.expanduser('~/.osb.yml')
        create_config_file({'foo': {'driver': 'ram'}}, filename)
        configs = utils.get_config_file()
        os.remove(filename)

    def test_bad_yaml(self):
        self.skipTest("Safe parsing enabled.")
        filename = tempfile.mktemp()
        with open(filename, 'w') as fd:
            fd.write('{1:2}')
        with self.assertRaises(errors.ConfigurationError):
            utils.get_config_file(filename)

    def test_no_file_found(self):
        with self.assertRaises(errors.ConfigurationError):
            configs = utils.get_config_file()


class GetDriverConfigTest(TestCase):
    def test_default_if_only_1_profile(self):
        create_config_file({'foo': {'driver': 'ram'}})
        driver_cfg = utils.get_driver_config()

    def test_profile_specified_but_not_found(self):
        filename = tempfile.mktemp()
        create_config_file({'foo': {'driver': 'ram'}}, filename)
        with self.assertRaises(errors.ConfigurationError):
            utils.get_driver_config(config_name='bar')
    
    def test_no_profile_found(self):
        create_config_file({})
        with self.assertRaises(errors.ConfigurationError):
            utils.get_driver_config(config_name='bar')


class GetDriverTest(TestCase):
    def test_func(self):
        config = {'driver': 'ram'}
        driver = utils.get_driver(config)

    def test_driver_not_found(self):
        config = {'driver': 'foo'}
        with self.assertRaises(Exception):
            driver = utils.get_driver(config)


class GetRandomNameTest(TestCase):
    def test_func(self):
        name = utils.get_random_name(20)
        self.assertEqual(len(name), 20)


class GetRandomContentTest(TestCase):
    def test_func(self):
        fd = utils.get_random_content(42)
        content = fd.read()
        self.assertEqual(len(content), 42)


class TimeItTest(TestCase):
    def test_output(self):
        def func():
            return True
        elapsed, output = utils.timeit(func)
        self.assertIsInstance(elapsed, float)
        self.assertIsInstance(output, bool)
        self.assertTrue(output)

    def test_elapsed(self):
        def func():
            time.sleep(1)
        elapsed, _ = utils.timeit(func)
        self.assertGreater(elapsed, 1)
        self.assertLess(elapsed, 1.1)



"""
.. note::
  This driver requires `azure-storage-blob`_.

Configuration
~~~~~~~~~~~~~

.. code-block:: yaml

  ---
  azure:
    driver: azure
    connect_str: <connect_str>

.. _azure-storage-blob: https://docs.microsoft.com/en-us/azure/storage/blobs/storage-quickstart-blobs-python
"""
from azure.storage import blob
from os_benchmark.drivers import base, errors

ACLS = {
    'public-read': 'container',
}


class Driver(base.RequestsMixin, base.BaseDriver):
    id = 'azure'
    client_kwargs = {}

    @property
    def client(self):
        if not hasattr(self, '_client'):
            self._client = blob.BlobServiceClient.from_connection_string(
                self.kwargs['connect_str'],
                **self.client_kwargs,
            )
        return self._client

    def setup(self, **kwargs):
        if 'multipart_chunksize' in kwargs:
            self.client_kwargs.update({
                'max_block_size': kwargs['multipart_chunksize'],
                'max_page_size': kwargs['multipart_chunksize'],
            })
        if 'multipart_threshold' in kwargs:
            self.client_kwargs['max_single_put_size'] = kwargs['multipart_threshold']
        self.logger.debug("Azure default config: %s", self.client_kwargs)
        self.client  # Avoid lazyness

    def list_buckets(self, **kwargs):
        containers = self.client.list_containers()
        return [{'id': c.name} for c in containers]

    def create_bucket(self, name, acl='public-read', **kwargs):
        bucket = self.client.create_container(
            name=name,
            public_access=ACLS.get(acl, None),
        )
        return {'id': name}

    def delete_bucket(self, bucket_id, **kwargs):
        try:
            self.client.delete_container(bucket_id)
        except blob.ResourceNotFoundError:
            pass

    def list_objects(self, bucket_id, **kwargs):
        client = self.client.get_container_client(bucket_id)
        blobs = client.list_blobs()
        return [b.name for b in blobs]

    def upload(self, bucket_id, name, content, max_concurrency=None,
               validate_content=False, **kwargs):
        max_concurrency = max_concurrency or base.MAX_CONCURRENCY
        client = self.client.get_container_client(bucket_id)
        client.upload_blob(
            name=name,
            data=content,
            max_concurrency=max_concurrency,
            timeout=self.read_timeout,
            validate_content=validate_content,
        )
        return {'name': name}

    def delete_object(self, bucket_id, name, **kwargs):
        client = self.client.get_container_client(bucket_id)
        client.delete_blob(name)

    def get_url(self, bucket_id, name, **kwargs):
        url = 'https://%s.blob.core.windows.net/%s/%s' % (
            self.client.account_name, bucket_id, name,
        )
        return url
